#include <assert.h>
#include <math.h>
#include "cairo-box.h"

enum
{
    NO_PROP,
    BOX_MODE,
    N_PROPERTIES
};
static GParamSpec *g_prop[N_PROPERTIES] = {0};

struct _DiyaCairoBox
{
    DiyaCairoContainer parent_object;
    guint mode;
};

G_DEFINE_FINAL_TYPE(DiyaCairoBox, diya_cairo_box, DIYA_TYPE_CAIRO_CONTAINER)

static void diya_cairo_box_dispose(GObject * object)
{
    g_debug("diya_cairo_box_dispose: %s", diya_object_to_string(object));
    G_OBJECT_CLASS(diya_cairo_box_parent_class)->dispose(object);
}

static void diya_cairo_box_init(DiyaCairoBox* self)
{
    self->mode = DIYA_CAIRO_BOX_HORIZONTAL;
}

static void diya_cairo_box_set_property(GObject *object, guint property_id, const GValue *value, GParamSpec *pspec)
{
    DiyaCairoBox * self = DIYA_CAIRO_BOX(object);
    switch (property_id)
    {
    case BOX_MODE:
        self->mode = g_value_get_uint(value);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
        break;
    }
}

static void diya_cairo_box_get_property(GObject *object, guint property_id, GValue *value, GParamSpec *pspec)
{
    DiyaCairoBox * self = DIYA_CAIRO_BOX(object);
    switch (property_id)
    {
    case BOX_MODE:
        g_value_set_uint(value, self->mode);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
        break;
    }
}

static gboolean cairo_box_draw(gpointer object, cairo_t * cairo)
{
    DIYA_CAIRO_WIDGET_CLASS(diya_cairo_box_parent_class)->draw(object, cairo);
    GList * children = DIYA_CAIRO_CONTAINER_GET_CLASS(object)->children(object);
    gpointer item;
    G_LIST_FOREACH(item, children)
    {
        DIYA_CAIRO_WIDGET_GET_CLASS(DIYA_CAIRO_WIDGET(item))->draw(item, cairo);
    }
    return TRUE;
}

static void cairo_box_update(gpointer object)
{
    diya_offset_t margin = {0};
    g_object_set(
        object,
        "halign", DIYA_CAIRO_WIDGET_ALIGN_START,
        "valign", DIYA_CAIRO_WIDGET_ALIGN_START,
        "margin", &margin,
        NULL
    );
    DIYA_CAIRO_WIDGET_CLASS(diya_cairo_box_parent_class)->update(object);

    assert(DIYA_IS_CAIRO_BOX(object));
    DiyaCairoBox * self = DIYA_CAIRO_BOX(object);
    GList * children = DIYA_CAIRO_CONTAINER_GET_CLASS(object)->children(object);
    if(!children)
    {
        return;
    }

    diya_rect_t space = diya_cairo_widget_get_extent(object);
    diya_offset_t * padding;
    diya_stroke_t * stroke;
    g_object_get(object,"padding", &padding, "stroke", &stroke, NULL);
    gint stroke_size = ceil(stroke->size);
    gpointer item;
    guint n_extended = 0;
    gint occupied_space = 2*stroke_size;
    guint vspace, hspace;
    guint cwidth, cheight;
    // explore all other children
    if(self->mode == DIYA_CAIRO_BOX_HORIZONTAL)
    {
        occupied_space += padding->left + padding->right;
    }
    else
    {
        occupied_space += padding->right + padding->bottom;
    }
    G_LIST_FOREACH(item, children)
    {
        g_object_get(item, "vspace", &vspace, "hspace", &hspace, NULL);
        DIYA_CAIRO_WIDGET_GET_CLASS(item)->content_size(item, &cwidth, &cheight);
        if(self->mode == DIYA_CAIRO_BOX_HORIZONTAL)
        {
            switch (hspace)
            {
            case DIYA_CAIRO_WIDGET_FILL_EXTEND:
                n_extended++;
                break;
            case DIYA_CAIRO_WIDGET_FILL_CONTENT:
            case DIYA_CAIRO_WIDGET_FILL_NONE:
                occupied_space += cwidth;
                break;
            default:
                break;
            }
        }
        else
        {
            switch (vspace)
            {
            case DIYA_CAIRO_WIDGET_FILL_EXTEND:
                n_extended++;
                break;
            case DIYA_CAIRO_WIDGET_FILL_CONTENT:
            case DIYA_CAIRO_WIDGET_FILL_NONE:
                occupied_space += cheight;
                break;
            default:
                break;
            }
        }
    }
    
    gint off_x = padding->left + stroke_size;
    gint off_y = padding->top + stroke_size;

    G_LIST_FOREACH(item, children)
    {
        g_object_get(item, "vspace", &vspace, "hspace", &hspace, NULL);
        DIYA_CAIRO_WIDGET_GET_CLASS(item)->content_size(item, &cwidth, &cheight);
        
        diya_rect_t rect = {0};
        rect.x = off_x;
        rect.y = off_y;
        if(self->mode == DIYA_CAIRO_BOX_HORIZONTAL)
        {
            rect.height = space.height - padding->top - padding->bottom - stroke_size*2;

            switch (hspace)
            {
            case DIYA_CAIRO_WIDGET_FILL_EXTEND:
                rect.width = (occupied_space >= space.width)?0:(space.width - occupied_space) / n_extended;
                off_x += rect.width;
                break;
            case DIYA_CAIRO_WIDGET_FILL_CONTENT:
            case DIYA_CAIRO_WIDGET_FILL_NONE:
                rect.width = cwidth;
                if(!DIYA_IS_CAIRO_BOX(item))
                {
                    rect.height = cheight;
                }
                off_x += cwidth;
                break;
            default:
                break;
            }
        }
        else
        {
            rect.width = space.width - padding->left - padding->right - stroke_size*2;
            switch (vspace)
            {
            case DIYA_CAIRO_WIDGET_FILL_EXTEND:
                rect.height = (occupied_space >= space.height)?0:(space.height - occupied_space) / n_extended;
                off_y +=  rect.height;
                break;
            case DIYA_CAIRO_WIDGET_FILL_CONTENT:
            case DIYA_CAIRO_WIDGET_FILL_NONE:
                rect.height = cheight;
                if(!DIYA_IS_CAIRO_BOX(item))
                {
                    rect.width = cwidth;
                }
                off_y += cheight;
                break;
            default:
                break;
            }
        }
        diya_rect_t * item_extent = DIYA_CAIRO_WIDGET_GET_CLASS(item)->extent(item);
        memcpy(item_extent, &rect, sizeof(rect));
    }

    G_LIST_FOREACH(item, children)
    {
        // back up vspace hspace
        g_object_get(item, "vspace", &vspace, "hspace", &hspace, NULL);
        g_object_set(item,
            "vspace", DIYA_CAIRO_WIDGET_FILL_EXTEND,
            "hspace", DIYA_CAIRO_WIDGET_FILL_EXTEND,
            NULL);
        DIYA_CAIRO_WIDGET_GET_CLASS(DIYA_CAIRO_WIDGET(item))->update(item);
        g_object_set(item,
            "vspace", vspace,
            "hspace", hspace,
            NULL);
    }
}

static diya_rect_t cairo_box_allocate_space(gpointer object, gpointer child)
{
    (void) object;
    return diya_cairo_widget_get_extent(child);
}

static void cairo_box_content_size(gpointer object, gint* width, gint * height)
{
    DiyaCairoBox * self = DIYA_CAIRO_BOX(object);
    *width = 0;
    *height = 0;
    gpointer item;
    GList * children = DIYA_CAIRO_CONTAINER_GET_CLASS(object)->children(object);
    diya_stroke_t * stroke;
    diya_offset_t * padding;
    g_object_get(object, "stroke", &stroke,"padding", &padding, NULL);
    gint stroke_size = ceil(stroke->size) * 2;
    G_LIST_FOREACH(item, children)
    {
        gint cw,ch;
        DIYA_CAIRO_WIDGET_GET_CLASS(item)->content_size(item,&cw,&ch);
        if(self->mode == DIYA_CAIRO_BOX_HORIZONTAL)
        {
            *width += cw;
            *height = MAX(*height,ch);
        }
        else
        {
            *height += ch;
            *width = MAX(*width,cw);
        }
    }
    *height += stroke_size + padding->bottom + padding->top;
    *width += stroke_size + padding->left + padding->right;
}

static void diya_cairo_box_class_init(DiyaCairoBoxClass * class)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(class);
    DIYA_CAIRO_WIDGET_SET_NAME(class, "DiyaCairoBox");
    
    DiyaCairoWidgetClass *wclass = DIYA_CAIRO_WIDGET_CLASS(class);
    
    wclass->draw = cairo_box_draw;
    wclass->update = cairo_box_update;
    wclass->alloc_space = cairo_box_allocate_space;
    wclass->content_size = cairo_box_content_size;

    gobject_class->dispose = diya_cairo_box_dispose;
    gobject_class->set_property = diya_cairo_box_set_property;
    gobject_class->get_property = diya_cairo_box_get_property;
    g_prop[BOX_MODE] = g_param_spec_uint(
        "mode", NULL,
        "Box mode", DIYA_CAIRO_BOX_HORIZONTAL,
        DIYA_CAIRO_BOX_VERTICAL, DIYA_CAIRO_BOX_HORIZONTAL,
        G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY);
    g_object_class_install_properties (gobject_class, N_PROPERTIES, g_prop);
}